from more_or_less.screen_dimensions import ScreenDimensions
import os
import shutil
import sys


class TerminalScreen(ScreenDimensions):
    '''
        Retrieves the dimensions of the current terminal window
    '''

    def __init__(self):
        pass

    def get_height(self):
        return self._get_height_and_width()[0]

    def get_width(self):
        return self._get_height_and_width()[1]

    def _get_height_and_width(self):
        columns, lines = shutil.get_terminal_size()
        return lines, columns



from .more_plugin import MorePlugin
from textwrap import dedent


class HelpPlugin(MorePlugin):
    '''
        Invoked when the user types 'h' or '?'.
        Prints the help text of all plugins
    '''

    def __init__(self):
        pass

    @property
    def keys(self):
        return ['h', '?']

    def build_page(self, page_builder, key_pressed, arguments):
        self._print_help(page_builder)
        return page_builder.build_next_page()

    def _print_help(self, page_builder):
        help_text = _HelpFormatter(plugins=page_builder.get_plugins()).format()
        page_builder.get_output().write(help_text)

    def _format_help(self):
        return self._help_format.format(help=self._format_plugin_help())

    def get_help(self):
        yield ('h or ?', 'Display this help text')


class _HelpFormatter(object):

    def __init__(self, plugins):
        self._plugins = plugins
        self._format_string = dedent('''\

            -------------------------------------------------------------------------------
            Most commands can optionally be preceded by an integer argument k.
            The default values are printed in brackets.
            A star (*) indicates the value of k becomes the new default.
            -------------------------------------------------------------------------------
            {}
            -------------------------------------------------------------------------------
        ''')

    def format(self):
        self._command_width = self._calculate_command_width()
        return self._format_string.format('\n'.join(self._iter_commands_help()))

    def _iter_commands_help(self):
        return (
            self._format_help_item(command, help_text)
            for plugin in self._plugins
            for command, help_text in plugin.get_help()
        )

    def _format_help_item(self, command, help_text):
        return '{command: <{width}}  {help_text}'.format(
            command=command,
            width=self._command_width,
            help_text=help_text)

    def _calculate_command_width(self):
        return max(len(command) for command, _ in self._get_help_items())

    def _get_help_items(self):
        for plugin in self._plugins:
            yield from plugin.get_help()



from abc import ABC, abstractmethod


class Page(ABC):
    '''
        A single output page.
        Text will be send to the page until 'is_full()' returns true.
        After that a new page will be used.

        When no more input is incoming (either because the page is full or the input ran out),
        'flush' will be called.
    '''

    @abstractmethod
    def is_full(self):
        # Returns 'true' if the page is full, or 'false' if more lines can be added.
        pass

    @abstractmethod
    def add_line(self, line):
        # Adds a line of output to this page.
        # The line is already terminated with a '\n'.
        pass

    def flush(self):
        # If any cleanup is required when a page is complete, it should be done here.
        pass



from . import more_plugins
from .buffered_input import BufferedInput
from .page_builder import PageBuilder, StopOutput
from .page_of_height import PageOfHeight
from .page_wrapper import PageWrapper
from .terminal_input import TerminalInput
from .terminal_screen import TerminalScreen
import sys


def MorePageBuilder(*args, **kwargs):
    '''
        A PageBuilder that is intended to work closely the way 'more' works.
        It supports the basic 'more' actions (one-more-page, n-more-lines, find-text).

        Extra actions can be installed as plugins, see more_plugins.py

        Constructor Arguments:
        ----------------------

        input: [type Input]
            If not specified we read input from stdin
        output: [type Output]
            If not specified we print output to stdout
        screen_dimensions: [type ScreenDimensions]
            If not specified we use the dimensions of the terminal window
        plugins: [type list of MorePlugin]
            The plugins to load. If not specified will fetch all plugins from more_plugins.py
    '''
    return PageWrapper(_MorePageBuilder(*args, **kwargs))


class _MorePageBuilder(PageBuilder):

    def __init__(self, input=None, output=None, screen_dimensions=None, plugins=None):
        self._screen_dimensions = screen_dimensions or TerminalScreen()
        self._output = output or sys.stdout
        self._input = BufferedInput(input or TerminalInput())

        self._plugins = plugins or more_plugins.get()
        self._action_handlers = _build_plugins_dictionary(self._plugins)

    def build_first_page(self):
        return PageOfHeight(height=self.get_page_height(), output=self._output)

    def build_next_page(self, message=None, arguments=None):
        try:
            return self._try_to_build_next_page(
                message or self.get_prompt_message(),
                arguments or {}
            )
        except KeyboardInterrupt:
            # Stop output on ctrl-c
            raise StopOutput

    def _try_to_build_next_page(self, message, arguments):
        while True:
            key_pressed = self._input.get_character(message)
            if key_pressed in self._action_handlers:
                handler = self._action_handlers[key_pressed]
                return handler.build_page(self, key_pressed=key_pressed, arguments=arguments)

    def get_plugins(self):
        return self._plugins

    def get_page_height(self):
        height_reserved_for_more_prompt = 1
        return self._screen_dimensions.get_height() - height_reserved_for_more_prompt

    def get_output(self):
        return self._output

    def get_input(self):
        ''' Returns the BufferedInput object used to get the user input '''
        return self._input

    def get_prompt_message(self):
        return '--More--'


def _build_plugins_dictionary(plugins):

    return {
        key: plugin
        for plugin in plugins
        for key in plugin.keys
    }



import sys

from more_or_less.screen_dimensions import ScreenDimensions


_HUGE = sys.maxsize


class FixedSizeScreen(ScreenDimensions):

    def __init__(self, height=_HUGE, width=_HUGE):
        self._height = height
        self._width = width

    def get_height(self):
        return self._height

    def get_width(self):
        return self._width



from .input import Input
import queue


class BufferedInput(Input):
    '''
        Input object that allows you to put back characters
        (which are then returned by the next call to 'get_character')
    '''

    def __init__(self, input):
        self._buffered_characters = queue.Queue()
        self._input = input

    def prompt(self, message):
        return self._input.prompt(message)

    def get_character(self, message):
        if not self._buffered_characters.empty():
            return self._buffered_characters.get()
        return self._input.get_character(message)

    def put_back(self, character):
        self._buffered_characters.put(character)



from .more_plugin import MorePlugin

_NON_REPEATABLE_COMMAND = '--Previous command can not be repeated--'


class RepeatPlugin(MorePlugin):
    '''
        Allows repeating of the previous command by pressing '.'.
        This only works if the previous command implements the RepeatableMixin.
    '''

    def __init__(self):
        self._last_page = None

    @property
    def keys(self):
        return ['.']

    def build_page(self, page_builder, key_pressed, arguments):
        # DO_NOT_PUSH test no previous page
        if _is_repeatable(self._last_page):
            return self._last_page.repeat()
        else:
            return page_builder.build_next_page(message=_NON_REPEATABLE_COMMAND)

    def wrap_page(self, page):
        self._last_page = page
        return page

    def get_help(self):
        yield ('.', 'Repeat previous command')


def _is_repeatable(page):
    return hasattr(page, 'repeat')



from .more_plugin import MorePlugin
from .page import Page
from .page_of_height import PageOfHeight
from .repeatable_mixin import RepeatableMixin
import re

_NO_PREVIOUS_REGULAR_EXPRESSION = '--No previous regular expression--'
_SKIPPING_MESSAGE = '...skipping\n'


class SearchPlugin(MorePlugin):
    ''' 
        Skips all output until a certain search pattern is found.
        Invoked when the user types '/'.
        The search can be repeated by pressing 'n'
    '''

    def __init__(self):
        self._pattern = None
        self._match_count = None

    @property
    def keys(self):
        return ['/', 'n']

    def build_page(self, page_builder, key_pressed, arguments):
        self._match_count = arguments.get('count', 1)

        if key_pressed == '/':
            return self._do_new_search(page_builder)
        elif key_pressed == 'n':
            return self._repeat_last_search(page_builder)
        else:
            assert False, 'Unexpected input key'

    def get_help(self):
        yield ('/<regular expression>', 'Search for kth occurrence of the regular expression [1]')
        yield ('n', 'Search for kth occurrence of the last regular expression [1]')

    def _do_new_search(self, page_builder):
        self._update_pattern(page_builder.get_input())
        return self._create_search_page(page_builder)

    def _repeat_last_search(self, page_builder):
        if self._pattern is None:
            return page_builder.build_next_page(message=_NO_PREVIOUS_REGULAR_EXPRESSION)
        else:
            return self._create_search_page(page_builder)

    def _create_search_page(self, page_builder):
        page_builder.get_output().write(_SKIPPING_MESSAGE)
        return SearchPage(
            pattern=self._pattern,
            next_page=self._create_full_page(page_builder),
            match_count=self._match_count,
        )

    def _create_full_page(self, page_builder):
        return PageOfHeight(
            height=page_builder.get_page_height(),
            output=page_builder.get_output())

    def _update_pattern(self, input):
        self._pattern = input.prompt('/')


class SearchPage(Page, RepeatableMixin):
    '''
        A page that suppresses all output until a given search pattern is found.
        After that it displays the passed in page
    '''

    def __init__(self, pattern, next_page, match_count):
        self.pattern = pattern
        self.next_page = next_page
        self._matcher = re.compile(pattern)
        self._actual_match_count = 0
        self.required_match_count = match_count

    def is_full(self):
        if self.has_match:
            return self.next_page.is_full()
        return False

    def add_line(self, line):
        self._match(line)

        if self.has_match:
            self.next_page.add_line(line)

    def _match(self, line):
        if self._matcher.search(line):
            self._actual_match_count = self._actual_match_count + 1

    def flush(self):
        if self.has_match:
            self.next_page.flush()

    def repeat(self):
        return SearchPage(self.pattern, self.next_page.repeat(), self.required_match_count)

    @property
    def has_match(self):
        return self._actual_match_count >= self.required_match_count



from .more_plugin import MorePlugin
from .page_builder import StopOutput


class QuitPlugin(MorePlugin):
    ''' 
        Invoked when the user types 'q'.
        Stops output 
     '''

    @property
    def keys(self):
        return ['q', 'Q']

    def build_page(self, page_builder, key_pressed, arguments):
        raise StopOutput()

    def get_help(self):
        yield ('q or Q or <interrupt>', 'Exit from more')



from .more_plugin import MorePlugin


class CountPlugin(MorePlugin):
    ''' 
        Invoked when the user types any number. 
        Adds a 'count' argument to the next action.
    '''

    def __init__(self):
        self._digits = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

    @property
    def keys(self):
        return self._digits

    def build_page(self, page_builder, key_pressed, arguments):
        arguments['count'] = self._get_count(page_builder, key_pressed)
        return page_builder.build_next_page(arguments=arguments)

    def get_help(self):
        return []

    def _get_count(self, page_builder, first_key):

        def iter_digits():
            # Read characters as long as the user enters digits
            key_pressed = first_key
            while key_pressed in self._digits:
                yield key_pressed
                key_pressed = input.get_character(prompt_message)
            input.put_back(key_pressed)

        input = page_builder.get_input()
        prompt_message = page_builder.get_prompt_message()

        return int(''.join(iter_digits()))



from abc import ABC, abstractmethod


class Output(ABC):
    '''
        Example API of what is expected from the 'output' object.
        This does not mean it must inherit from this.

        Note that any 'file' object matches this API,
        so files can natively be used as output.
    '''

    @abstractmethod
    def write(self, text):
        pass

    @abstractmethod
    def flush(self):
        pass



from abc import ABC, abstractproperty, abstractmethod


class MorePlugin(ABC):
    '''
        A plugin that represents an extra action the user can take on the 'more' prompt.

    '''

    @abstractproperty
    def keys(self):
        ''' Returns a list of the keys the user has to enter to trigger this action. '''
        pass

    @abstractmethod
    def build_page(self, page_builder, key_pressed, arguments):
        ''' 
            Called when the user pressed one of the keys to trigger this action.

            Arguments:
            ----------

            page_builder:
                The MorePageBuilder instance.
            key_pressed:
                The key the user pressed to trigger this action.
            arguments:
                A dictionary of arguments the user entered on this line before triggering
                this action.
                By default, the only value that can be in there is 'count',
                which will be set if the user entered a number before your action.

                For example, if the user entered '10 ' then the '<space>' action is triggered
                with argument {'count': 10}.
        '''
        pass

    def wrap_page(self, page):
        ''' 
            Called when a new page is created. 
            Gives the plugin to return a wrapper page that can be used to modify/register
            _every_ line, including the ones that are suppressed by other plugins.

            Example usage is counting all the outputted lines.

            Must return a 'Page'. Implementing this method is optional.
        '''
        return page

    @abstractmethod
    def get_help(self):
        '''
            Returns an iterator over 'command', 'help-text' tuples that describe how to use
            this plugin.
            Example:
            yield (' ', 'Display next line of text')
        '''
        pass



from .more_plugin import MorePlugin
from .page_of_height import PageOfHeight


class OnePagePlugin(MorePlugin):
    ''' 
        Displays a new full output page.
        Invoked when the user types ' '. 
    '''

    @property
    def keys(self):
        return [' ']

    def build_page(self, page_builder, key_pressed, arguments):
        page_height = arguments.get('count', page_builder.get_page_height())
        return PageOfHeight(height=page_height, output=page_builder.get_output())

    def get_help(self):
        yield ('<space>', 'Display next k lines of text [current screen size]')



from .page import Page
from .repeatable_mixin import RepeatableMixin


class PageOfHeight(Page, RepeatableMixin):
    '''
        A page that accepts a given number of lines.
        Every line will be forwarded to the given 'output' object
        (of type output.Output).
    '''

    def __init__(self, height, output):
        self.output = output
        self._remaining_lines = height
        self.height = height

    def is_full(self):
        return self._remaining_lines == 0

    def add_line(self, line):
        self._remaining_lines = self._remaining_lines - 1
        self.output.write(line)

    def flush(self):
        self.output.flush()

    def repeat(self):
        return PageOfHeight(self.height, self.output)



from .page import Page
from abc import ABC, abstractmethod


class WrappedPage(Page, ABC):
    '''
        Basic class that can be derived from if you need to
        create a Page that
           - records all printed lines
           - can change the printed lines before forwarding.

    '''

    def __init__(self, wrapped_page):
        self.wrapped_page = wrapped_page

    def is_full(self):
        return self.wrapped_page.is_full()

    def add_line(self, line):
        new_line = self.on_add_line(line)
        return self.wrapped_page.add_line(new_line)

    def flush(self):
        return self.wrapped_page.flush()

    @abstractmethod
    def on_add_line(self, line):
        ''' Called with every line. Returns the modified version of the line '''
        pass

    def __getattr__(self, name):
        return getattr(self.wrapped_page, name)



from .count_plugin import CountPlugin
from .help_plugin import HelpPlugin
from .line_count_plugin import LineCountPlugin
from .more_plugin import MorePlugin
from .one_line_plugin import OneLinePlugin
from .one_page_plugin import OnePagePlugin
from .quit_plugin import QuitPlugin
from .repeat_plugin import RepeatPlugin
from .search_plugin import SearchPlugin

__plugins = [
    CountPlugin,
    OnePagePlugin,
    OneLinePlugin,
    QuitPlugin,
    LineCountPlugin,
    SearchPlugin,
    RepeatPlugin,
    HelpPlugin,
]


def add_plugin(handler):
    '''
        Adds a MorePlugin class. 
        This allows you to add extra keys that the user can enter on the '--More--' prompt.

        Note that you must pass a callable in here that, when invoked, returns a MorePlugin.
    '''
    assert callable(handler)
    assert isinstance(handler(), MorePlugin)
    __plugins.append(handler)


def remove_plugin(handler):
    '''
        Removes a MorePlugin class. 
    '''
    __plugins.remove(handler)


def get():
    return [plugin() for plugin in __plugins]



#!python
from .more_page_builder import MorePageBuilder
from .page_builder import StopOutput
import queue
import threading


# Signal to send to the input queue when there is no more input
END_OF_INPUT = None
# Return code if output was interrupted by the user (e.g. the user pressed ctrl+c)
OUTPUT_STOPPED = 'OUTPUT_STOPPED'


def paginate(
        input,
        output=None,
        prompt=None,
        screen_dimensions=None,
        plugins=None,
        page_builder=None,
        asynchronous=False):
    '''
        Paginates the input, similar to how 'more' works in bash.

        Reads from input until the output window is full.
        Then prompts the user for an action before reading more input.

        Pseudo-logic:
        -------------

        page = page_builder.build_first_page()
        for line in <input-lines>:
            if page.is_full():
                page.flush()
                page = page_builder.build_next_page()
            page.add_line(line)

        Arguments:
        ----------

        input: [type iterable or Queue]
            The input text that should be paginated.
            This must either be an iterable over text (e.g. a list or a file), or an instance of queue.Queue.

            It is not required that each returned string is a complete line. 
            The paginator will combine incomplete lines until a '\n' is encountered.

            If it is a queue.Queue, you must pass 'END_OF_INPUT' into the queue when no more input is expected.
            This will flush the final incomplete line (if any) to the output.
            Note that you can NOT use queue.join() to detect all input has been processed
            (as that just raises issues if the user decides to abort the output halfway through).
            Instead, if you use 'asynchronous=True' you can join the returned context.

        output: [type Output]
            If not specified we print output to stdout

        prompt: [type Input]
            Used when prompting the user for actions.
            Defaults to reading from stdin.

        screen_dimensions: [type ScreenDimensions]
            Used to know the height of the output window 
            (which is used to decide how many lines to print before we consider a page 'full').
            Defaults to using the dimensions of the terminal window.

        plugins: [type list of MorePlugin]
            The plugins to load. These plugins decide what actions are available on the 'more' prompt.
            If not specified will fetch all plugins from more_plugins.py

        asynchronous: [type bool] 
            If true the 'paginate' call will return instantly and run asynchronously.
            In this case a context is returned on which you can call 'context.join([timeout])' 
            to block until all lines are sent to the output.

        page_builder: [type PageBuilder]
            The object that will create the output pages whenever a page is full.
            Must be an instance of 'PageBuilder'.
            If specified we ignore the values of output, prompt, screen_dimensions and plugins.


        Returns:
        --------

        A joinable 'context' if asynchronous is True
        OUTPUT_STOPPED if the user stopped the output (for example using ctrl+c)

    '''

    page_builder = page_builder or MorePageBuilder(
        input=prompt,
        output=output,
        screen_dimensions=screen_dimensions,
        plugins=plugins)

    if asynchronous:
        thread = threading.Thread(
            target=paginate,
            kwargs={
                'input': input,
                'page_builder': page_builder,
                'asynchronous': False,
            },
        )
        thread.start()
        return thread

    paginator = Paginator(page_builder)
    if isinstance(input, queue.Queue):
        return paginator.paginate_from_queue(input)
    else:
        return paginator.paginate(input)


class Paginator(object):
    '''
        Paginates given input text, similar to how 'more' works in bash.

        See help of 'paginate' for a more detailed description of the behavior.

        There are 3 ways to send input text:
            - pass an iterable to self.paginate.
            - pass a queue to self.paginate_from_queue.
            - call 'add_text' repeatedly until all text has been sent in, then call 'flush_incomplete_line'.

        Each of these methods returns 'OUTPUT_STOPPED' if the user stopped the output (for example using ctrl+c)
    '''

    def __init__(self, page_builder):
        self._page_builder = page_builder
        self._lines = _LineCollector()

        self._page = self._page_builder.build_first_page()

    def paginate(self, iterable):
        '''
            Iterates over the iterable, and paginates all the text it returns
        '''
        try:
            for text in iterable:
                self._try_to_add_text(text)

            self.flush_incomplete_line()
        except StopOutput:
            return OUTPUT_STOPPED

    def paginate_from_queue(self, input_queue):
        '''
            Iterates over the queue, and paginates all the text it returns.
            Stops paginating when END_OF_INPUT is encountered on the queue.
        '''
        return self.paginate(QueueIterator(input_queue))

    def add_text(self, input_text):
        '''
            Splits the input_text into lines, and paginates them.
            Can be called multiple times.

            When you're done you must call 'flush_incomplete_line'
            to ensure the last incomplete input line is sent to the output.
        '''
        try:
            self._try_to_add_text(input_text)
        except StopOutput:
            return OUTPUT_STOPPED

    def _try_to_add_text(self, input_text):
        self._lines.add(input_text)

        for line in self._lines.pop_complete_lines():
            self._paginate_and_print_text(line)

    def flush_incomplete_line(self):
        try:
            self._try_to_flush_incomplete_line()
        except StopOutput:
            return OUTPUT_STOPPED

    def _try_to_flush_incomplete_line(self):
        if len(self._lines.incomplete_line):
            self._paginate_and_print_text(self._lines.pop_incomplete_line())
        self._page.flush()

    def _paginate_and_print_text(self, text):
        if self._page.is_full():
            self._start_new_page()
        self._output_text(text)

    def _start_new_page(self):
        self._page.flush()
        self._page = self._page_builder.build_next_page()

    def _output_text(self, text):
        self._page.add_line(text)


class _LineCollector(object):
    '''
        Collects the input text and allows us to walk over the complete lines only.
        example:
             self.add('first ')
             self.add('line \nsecond line\n')
             self.add('incomplete final line')

             self.pop_complete_lines() <-- returns ['first line', 'second line']
             self.pop_incomplete_line() <-- returns 'incomplete final line'

    '''

    def __init__(self):
        self._complete_lines = []
        self.incomplete_line = ''

    def add(self, text):
        assert isinstance(text, str), 'expected str got {}'.format(text.__class__)

        unprocessed_text = self.incomplete_line + text
        complete_lines, incomplete_line = self._split_lines(unprocessed_text)

        self._complete_lines += complete_lines
        self.incomplete_line = incomplete_line

    def pop_complete_lines(self):
        try:
            return self._complete_lines
        finally:
            self._complete_lines = []

    def pop_incomplete_line(self):
        try:
            return self.incomplete_line
        finally:
            self.incomplete_line = ''

    def _split_lines(self, text):
        lines = text.splitlines(True)

        if self._has_incomplete_line(lines):
            complete_lines = lines[:-1]
            incomplete_line = lines[-1]
        else:
            complete_lines = lines
            incomplete_line = ''

        return (complete_lines, incomplete_line)

    def _has_incomplete_line(self, lines):
        return len(lines) and not lines[-1].endswith('\n')


def _make_callable(value):
    if not callable(value):
        return lambda: value
    else:
        return value


class QueueIterator(object):
    ''' 
        Iterates over a queue, until END_OF_INPUT is encountered 
    '''

    def __init__(self, queue):
        self._queue = queue

    def __iter__(self):
        return self

    def __next__(self):
        text = self._queue.get()
        if text is END_OF_INPUT:
            raise StopIteration
        return text



from abc import ABC, abstractmethod


class Input(ABC):
    '''
        Input object that supports either reading a single character,
        or a complete input line.
    '''

    @abstractmethod
    def prompt(self, message):
        pass

    @abstractmethod
    def get_character(self, message):
        pass



from abc import ABC, abstractmethod


class RepeatableMixin(ABC):
    '''
       If your Page inherits from this,
       the action it represents can be repeated by pressing '.'
    '''

    @abstractmethod
    def repeat(self):
        ''' 
            Returns an instance of a new version of this page.
            For example, a repeat of a page of height X
            returns a page of height X, on which 0 lines of output have been printed.
    '''



from .more_plugin import MorePlugin
from .page_of_height import PageOfHeight


class OneLinePlugin(MorePlugin):
    ''' 
        Displays one more output line.
        Invoked when the user types '<enter>'. 
    '''

    def __init__(self):
        self._page_height = 1

    @property
    def keys(self):
        return ['\r', '\n']

    def build_page(self, page_builder, key_pressed, arguments):
        self._update_page_height(arguments)
        return PageOfHeight(height=self._page_height, output=page_builder.get_output())

    def get_help(self):
        yield ('<return>', 'Display next k lines of text [{}]*'.format(self._page_height))

    def _update_page_height(self, arguments):
        self._page_height = arguments.get('count', self._page_height)



from .page_builder import PageBuilder


class PageWrapper(PageBuilder):
    '''
        Wrapper around the 'build_[first|next]_page' methods that calls all the plugins.wrap_page
        methods.
    '''

    def __init__(self, actual_page_builder):
        self._actual_page_builder = actual_page_builder

    def build_first_page(self):
        return self._wrap_page(self._actual_page_builder.build_first_page())

    def build_next_page(self):
        return self._wrap_page(self._actual_page_builder.build_next_page())

    def __getattr__(self, name):
        return getattr(self._actual_page_builder, name)

    def _wrap_page(self, page):
        for plugin in self.get_plugins():
            page = plugin.wrap_page(page)
        return page



from .input import Input
import os
import sys
import termios
import tty


# http://ascii-table.com/ansi-escape-sequences.php
_CLEAR_TEXT_ON_LINE = '\x1b[2K'
_MOVE_CURSOR_TO_START_OF_LINE = '\x1b[1000D'  # Actually moves the cursor 1000 characters to the left
ERASE_LINE = _CLEAR_TEXT_ON_LINE + _MOVE_CURSOR_TO_START_OF_LINE


class TerminalInput(Input):
    '''
        Input object that gets the input from the user terminal.
    '''

    def __init__(self):
        self._input = _get_interactive_input_stream()
        self._output = sys.stdout

    def prompt(self, message):
        self._print_prompt(message)
        return self._input.readline().rstrip('\n')

    def get_character(self, message):
        try:
            self._print_prompt(message)
            return _read_character(self._input)
        finally:
            self._erase_input_prompt()

    def _print_prompt(self, message):
        print(message, end='', file=self._output, flush=True)

    def _erase_input_prompt(self):
        self._output.write(ERASE_LINE)


def _get_interactive_input_stream():
    if sys.stdin.isatty():
        return sys.stdin
    else:
        '''
            We come here if the stdin is read from a file/pipe.
            In this case, that is probably the content we want to paginate,
            so we will open a new interactive input-stream.

            See https://stackoverflow.com/questions/46441623/get-user-input-while-reading-stdin-from-a-pipe/46445955#46445955
        '''
        return _create_interactive_input_stream()


def _create_interactive_input_stream():
    if sys.platform in ('win32', 'cygwin'):

        raise NotImplemented('Reading interactive input while stdin is not a tty is not supported on Windows')
    try:
        return open('/dev/tty', 'r')
    except FileNotFoundError:
        raise AssertionError('Terminal is unavailable for interactive input (could not open /dev/tty).')


def _read_character(file):
    '''' reads a single character. Does not wait for 'enter' '''
    fd = file.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setcbreak(fd)
        return os.read(fd, 1).decode('utf-8')
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)



from .fixed_size_screen import FixedSizeScreen
from .input import Input
from .line_count_plugin import LineCountPlugin
from .more_page_builder import MorePageBuilder
from .more_plugin import MorePlugin
from .more_plugins import add_plugin, remove_plugin
from .output import Output
from .page import Page
from .page_builder import PageBuilder, StopOutput
from .page_of_height import PageOfHeight
from .paginator import Paginator, paginate, END_OF_INPUT, OUTPUT_STOPPED
from .repeatable_mixin import RepeatableMixin
from .screen_dimensions import ScreenDimensions
from .wrapped_page import WrappedPage


__all__ = [
    END_OF_INPUT,
    FixedSizeScreen,
    Input,
    LineCountPlugin,
    MorePageBuilder,
    MorePlugin,
    OUTPUT_STOPPED,
    Output,
    Page,
    PageBuilder,
    PageOfHeight,
    Paginator,
    RepeatableMixin,
    ScreenDimensions,
    StopOutput,
    WrappedPage,
    add_plugin,
    paginate,
    remove_plugin,
]



#!python
from setuptools import setup, find_packages

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='more-or-less',
    version="0.1.5",
    description="Library for adding 'more' like paging functionality to your Python application",
    long_description=long_description,
    author="Jeroen Dhollander",
    url='https://github.com/jeroen-dhollander/python-more-or-less',
    packages=find_packages(),
    include_package_data=True,
)



from .more_plugin import MorePlugin
from .wrapped_page import WrappedPage


class LineCountPlugin(MorePlugin):
    ''' 
        Prints the current line number on '='.
        Toggles prefixing all lines with the line number on 'l'.
    '''

    def __init__(self):
        self.line_numbers_enabled = False
        self.line_count = 0

    @property
    def keys(self):
        return ['=', 'l', 'L']

    def get_help(self):
        yield ('=', 'Display current line number')
        yield ('l', f'Toggle printing line number on every line [currently {self._format_enabled()}]')

    def build_page(self, page_builder, key_pressed, arguments):
        if key_pressed == '=':
            return page_builder.build_next_page(message=f'--{self.line_count}--')
        elif key_pressed == 'l':
            self.line_numbers_enabled = not self.line_numbers_enabled
            return page_builder.build_next_page(message=f'--Line numbers are now {self._format_enabled()}--')
        else:
            assert False, 'Unexpected key event'

    def wrap_page(self, page):
        return _LineCounter(self, page)

    def _format_enabled(self):
        return {
            True: 'enabled',
            False: 'disabled',
        }.get(self.line_numbers_enabled)


class _LineCounter(WrappedPage):

    def __init__(self, line_count_plugin, wrapped_page):
        super().__init__(wrapped_page)
        self._plugin = line_count_plugin

    def on_add_line(self, line):
        self._bump_line_count()
        if self.must_add_line_number():
            return self.add_line_number(line)
        else:
            return line

    def _bump_line_count(self):
        self._plugin.line_count = self._plugin.line_count + 1

    def must_add_line_number(self):
        return self._plugin.line_numbers_enabled

    def add_line_number(self, text):
        return f'{self._plugin.line_count}: {text}'



from abc import ABC, abstractmethod


class StopOutput(Exception):
    '''
        Exception raised to indicate there is no next page.
    '''


class PageBuilder(ABC):
    '''
        Builds the output pages (of type Page) used in the paginator.

        When the paginator start, it will call build_first_page.
        When that page is full, it will call build_next_page.
        This is repeated until there is no more input.

    '''

    @abstractmethod
    def build_first_page(self):
        pass

    @abstractmethod
    def build_next_page(self):
        '''
            raise StopOutput to indicate there is no next page
        '''
        pass
